#include "biquad.h"
#include <vector>
#include <iostream>
#include <iomanip>
#include "bad_thing.h"

using namespace std;

//  c0, c1, c2 are the coefficients in the numerator
// (1), d1, d2 are the coefficients in the denominator

double biquad::operator()(const double Vin){
// Direct Form II
// ws stands for "weighted sum"
    double q1(ws1);             // save previous values
    double q2(ws2);
    double Vout = c0*Vin           + q1;
    ws1         = c1*Vin - d1*Vout + q2;
    ws2         = c2*Vin - d2*Vout;
    return Vout;
}

// transfer function
C biquad::xfunc(const C z){
    C nume = (c0*z + c1)*z + c2;
    C denom = (z + d1)*z + d2;
    return nume / denom;
}

void biquad::normalize(){
  double norm = xfunc(1.).real();
  c0 /= norm;
  c1 /= norm;
  c2 /= norm;
}

// angle is the angle in the z-plane, in radians
// i.e. 2 * pi * notch_freq / sampling_freq
// i.e. 1 * pi * notch_freq / Nyquist_freq
notch::notch(const double angle)
: biquad(1)
{
//  C topzero(cos(angle), sin(angle));
  double zero_re(cos(angle));
  double zero_im(sin(angle));
  c0 = 1;
  c1 = -2.*zero_re;
  c2 = 1;

  double pole_re(1. - .5*sqrt(1. - zero_re));
  double pole_im(zero_im/3.);
  d1 = -2 * pole_re;
  d2 = pole_re*pole_re + pole_im*pole_im;

  normalize();
}

void biquad::dump_coeff(){
  cout        << setw(8) << c0
       << " " << setw(8) << c1
       << " " << setw(8) << c2 << endl;
  cout        << setw(8)
       << " " << "(1)"
       << " " << setw(8) << d1
       << " " << setw(8) << d2 << endl;
}

#ifdef BUTTERWORTH

// Two-pole Butterworth bandpass filter.  Derived class.
// The corner is specified as an angle in the z-plane, in radians
// i.e. 2 * pi * corner_freq / sampling_freq
// i.e. 1 * pi * corner_freq / Nyquist_freq

void butterworth_2_lp::set_corner(const double angle){

// The frequency argument required for the denominator
// is not in radians, but rather twice that; i.e.
// it is frequency relative to Nyquist rate, not sampling rate.
// Don't ask me why:
  vector<double> ccof = ccof_bwlp(2);  /* independent of frequency */
  vector<double> dcof = dcof_bwlp(2, angle/M_PI);
  c0 = ccof[0];
  c1 = ccof[1];
  c2 = ccof[2];
  d1 = dcof[1];
  d2 = dcof[2];

#if 0
  for (unsigned int ii = 0; ii < ccof.size(); ii++){
    cout << setw(8) << ccof[ii] << " ";
  }
  cout << " <-- ccof" << endl;

  for (unsigned int ii = 0; ii < dcof.size(); ii++){
    cout << setw(8) << dcof[ii] << " ";
  }
  cout << " <-- dcof" << endl;
#endif
  normalize();
}

// Default constructor, required if you are going to make
// a vector<> of these things:
butterworth_2_lp::butterworth_2_lp()
: biquad(1)
{}

// fancy constructor
butterworth_2_lp::butterworth_2_lp(const double angle)
: biquad(1)
{
  set_corner(angle);
}

#endif


#if 0
// typical usage:
biquad butterworth(1.,          2.,                     1.,
                          -1.99911142347079540116, 0.99911181807963833634);
#endif
