// about 110 ns per byte on my laptop
// using ./sha1-test -times 100000

/*

Test driver in C++ for SHA-1 in C
By Steve Reid <steve@edmweb.com>
100% Public Domain

Test Vectors (from FIPS PUB 180-1)
"abc"
  A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D
"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
  84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1
A million repetitions of "a"
  34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
*/

using namespace std;
#include <iostream>
#include <iomanip>
#include <cmath>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>	/* for open */
#include <sys/stat.h>	/* for open */
#include <fcntl.h>	/* for open */
#include <unistd.h>	/* for read */
#include <string>
#include "sha1.h"
#include "sha1-more.h"
#include "utils.h"      /* for getTime */

void usage(int err){
  (err ? cerr : cout) <<
"Test driver for SHA-1\n"
"Based on code from Steve Reid  <steve@edmweb.com>\n"
"Usage:\n"
"	sha1-test 'filename'\n"
"	sha1-test [options]\n"
" Options include:\n"
" -	## reads stdin\n"
" -abc	## test vector, length 3\n"
"   should be:   A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D\n"
" -abcd	## test vector, length 56\n"
"   should be:   84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1\n"
" -aaa	## test vector, length 1,000,000\n"
"   should be:   34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F\n"
" -times nnnn     ## perform the hash nnnn times and report statistics.\n"
" -symspace nnnn  ## set size of symbol space, e.g. 256 for single byte.\n"
" -margin   nnnn  ## set margin (in bits);  10 is reasonable.\n"
" -show     nnn   ## tell hashem to display the first nnn rows of the histogram.\n"
" -hashem         ## compute entropy, using specified symspace and margin\n"
"\n"
" Example:\n"
"    ./sha1-test -margin 10 -symspace 256 -show 20 -hashem\n"
" Result:\n"
"  Entropy/bits: 7.99935  density: 99.9919 %  deficit: 0.000651495  scaled: 0.667131\n"
"\n"
"    ./sha1-test -margin 15 -symspace 256 -show 20 -hashem\n"
" Result:\n"
"  Entropy/bits: 7.99998  density: 99.9997 %  deficit: 2.34573e-05  scaled: 0.768649\n"
"  (runtime less than 1 minute)\n"
;
exit(err);
}


struct Histo{
  int symbolspacesiz;         // e.g. 256 for 8-bit symbols
  int margin;                 // measured in bits
  int show_histo;
  Histo()
  : symbolspacesiz(256),
    margin(0),
    show_histo(0)
  {}
  void doit(){
    SHA1_CTX context;
    unsigned char digest[20];
    const int nwds(128);      // like the input pool in /dev/random
    const int nbytes(nwds*sizeof(int));
    assert(nbytes == 512);
    unsigned int inbuf[nwds];
    const int histosiz = 256*256;
    unsigned int histo[histosiz];
    assert(symbolspacesiz <= histosiz);
    int todo = symbolspacesiz<<margin;
  // Suppose symbolspacesiz is 6553 and margin = 10;
  // Then todo is 67,108,864.
  // This takes 7.5 minutes to compute on my laptop (October 2013).

// Initialize the histogram to zero:
    for (int ii = 0; ii < histosiz; ii++) histo[ii] = 0;
// Ditto for the input buffer:
    for (int ii = 0; ii < nwds; ii++) inbuf[ii] = 0;

// Loop over all possible inputs,
// within a certain class of low-entropy inputs,
// formed by a simple counter.
// Histogram the resulting output symbols.
// This is essentially a Monte Carlo integration over
// the probability measure.  You can think of probability
// measure as an indestructible fluid that flows through the
// hash function, from input to output.  For details on what
// a probability measure is, see
//    http://www.av8n.com/physics/probability-intro.htm#sec-prob-measure
// In the simple no-margin case, we assign equal weight to
// each of the possible inputs.

// The _margin_ allows us to work in a bigger space, a
// product space: the outer product of
//    (all possible symbols) (times) (all possible margins)
// In the presence of a margin, the integral assigns
// equal value to every point in this product space.
// In some rough sense, the margin is similar to the seed of
// a PRNG.
    for (int ii = 0; ii < todo; ii++) {
      inbuf[0] = ii;
      SHA1_Init(&context);
      SHA1_Update(&context, inbuf, nbytes);
      SHA1_Final(digest, &context);
      int symbol = digest[0] + digest[1]*256;
      symbol = symbol % symbolspacesiz;
      histo[symbol]++;
    }
    unsigned int * hptr(histo);

// Optionally, show the entries in the histogram.
    if (show_histo) {
      int maxrow = histosiz/16;
      if (maxrow > show_histo) maxrow = show_histo;
      for (int ii = 0; ii < maxrow; ii++){
        for (int jj = 0; jj < 16; jj++) {
          if (jj) cout << " ";
          int nnn = *hptr++;
          cout << setw(4) << nnn;
        }
        cout << endl;
      }
    }

// Make sure we didn't lose any probability:
    int checksum = 0;
    for (int ii = 0; ii < histosiz; ii++) {
      checksum += histo[ii];
    }
    if (todo != checksum){
      cerr  << "Egads! todo: " << todo
            << "  did: " << checksum
            << endl;
      exit(1);
    }

// The histogram *is* a probability distribution,
// namely a frequency, and as such it has an entropy.
// Calculate the entropy directly from the definition,
// entropy = Sum Pi log (1/Pi)
    double entropy(0);
    for (int ii = 0; ii < histosiz; ii++) {
      double Pi = histo[ii] / double(todo);
      if (Pi) entropy -= Pi * log2(Pi);
    }

// Display the results:
    double smax = log2(symbolspacesiz);
    double deficit = smax - entropy;
    cout << "Entropy/bits: " << entropy
          << "  density: " << entropy/smax*100. << " %"
          << "  deficit: " << deficit
          << "  scaled: " << deficit * pow(2, margin)
          << endl;
  }
};

void dump(Digest& digest) {
  for (int i = 0; i < 5; i++) {
      if (i) fprintf(stdout, " ");
      fprintf(stdout, "%08X", htonl(digest.word[i]));
  }
}

void finish(SHA1_CTX &context, char const * shouldbe = 0){
  Digest digest;
  SHA1_Final(digest.byte, &context);
  dump(digest);
  if (shouldbe) {
    if (digest == Digest(shouldbe)){
      cout << " --> Check OK." << endl;
      return;
    } else {
      cout << " --> Check failed!!!" << endl;
      exit(1);
    }
  } else {
    cout << endl;
  }
}

void check_abc(){
  SHA1_CTX context;
  SHA1_Init(&context);
  SHA1_Update(&context, "abc", 3);
  finish(context, "A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D");
}

void check_abcd(){
  SHA1_CTX context;
  SHA1_Init(&context);
  SHA1_Update(&context,
"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", 56);
  finish(context, "84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1");
}

void check_aaa(){
  SHA1_CTX context;
  SHA1_Init(&context);
  for (int ii = 0; ii < 20000; ii++) SHA1_Update(&context,
"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", 50);
  finish(context, "34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F");
}

void check_times(int const count){
  double start = getTime();
  Digest digest(0);
  for (int ii = 0; ii < count; ii++) {
    SHA1_CTX context;
    SHA1_Init(&context);
    SHA1_Update(&context, "abc", 3);
    SHA1_Update(&context, digest.byte, sizeof(digest.byte));
    SHA1_Final(digest.byte, &context);
  }
  double elapsed = getTime() - start;
  cout << "count: " << count
    << "  elapsed: " << elapsed
    << "  each: " << elapsed/count
    << "  i.e. " << 1e9*elapsed/count/20. << " ns per byte"
    << endl;
}

void check_file(const string fname){
  SHA1_CTX context;
  Digest digest;
  int fd;
  if (fname == "-") fd = fileno(stdin);
  else fd = open(fname.c_str(), O_RDONLY);
  if (fd < 0) {
    cerr <<"Unable to open file '" << fname << "'" << endl;
    exit(1);
  }
  int i;
  unsigned char buffer[16384];

  SHA1_Init(&context);
  while (1) {
      i = read(fd, buffer, 16384);
      if (i <= 0) break;
      SHA1_Update(&context, buffer, i);
  }
  finish(context);
}

int main(int argc, char** argv){
  string progname(argv[0]); argc--; argv++;
  string arg;
  Histo histo;
  while (argc) {
    arg = *argv++;  argc--;
    if (0) {}
    else if (arg == "-margin") {
      histo.margin = atoi(*argv++); argc--;
    }
    else if (arg == "-symspace") {
      histo.symbolspacesiz = atoi(*argv++); argc--;
    }
    else if (arg == "-show") {
      histo.show_histo = atoi(*argv++); argc--;
    }
    else if (arg == "-hashem") histo.doit();
    else if (arg == "-times") {
      int times = atoi(*argv++); argc--;
      check_times(times);
    }
    else if (arg == "-h") usage(0);
    else if (arg == "-abc") check_abc();
    else if (arg == "-abcd") check_abcd();
    else if (arg == "-aaa") check_aaa();
    else if (arg == "-") check_file(arg);
    else if (arg[0] == '-') {
      cerr << "Unrecognized option '" << arg << "'" << endl;
      exit(1);
    }
    else check_file(arg);
  }
  exit(0);
}
